#!/usr/bin/perl
use strict;
use Getopt::Long;

use Data::Dumper qw(DumperX);
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse = 1;

GetOptions("in=s" => \ my $FileName,
           "list=s" => \ my $list,
           "edge=i" => \ my $c,
           "load=i" => \ my $d,
           "out=s" => \ my $out,
           "debug=i" => \ my $debug
           );

#my $debug;
if(!defined($debug)) {
	$debug = 0;
}

my $timingMult = 1e12;
my $capMult = 1e15;

my $CellName;
my $spiceFile;
my @inFiles;
if(defined($FileName)) {
     $CellName = $FileName;
     $CellName =~ s/\..*//;
     $spiceFile = $CellName . ".spi";
     push(@inFiles, $FileName);
} elsif(defined($list)) {
     if(-f $list) {
          open(LIST, $list) || die "Cannot open $list for read, $!\n";
          while(<LIST>) {
               chomp;
               if($_) {
                    push(@inFiles, $_);
               }
          }
          close(LIST);
     }
}
               
my $LibName;
if(!defined($out)) {
     if(defined($FileName)) {
          $LibName = $CellName;
          $out = $LibName . ".lib";
     } else {
          $LibName = $list;
          $out = $LibName . ".lib";
     }
}

my @arcKeywords = ( "type",
		    "input",
		    "output",
		    "clock",
		    "tie",
		    "drive",
		    "ic",
		    "state",
		    "arc",
		    "endarc",
		    "delay",
		    "setup",
		    "hold",
		    "rise",
		    "fall",
		    "high",
		    "low",
		    "pin",
		    "endpin",
		    "edgerates",
		    "outloads",
		    "direction",
		    "clk_edge_rate",
		    "pincap",
		    "syn_func",
              "clock_pin_true"
);

my @spiceKeywords = ( "xcut",
		      "vp",
		      "vss",
		      "out_load",
		      "edge_rate",
		      "start_time",
		      "setup_",
		      "clk_edge_rate",
		      "state_volt"
);

#make SURE you are using the right version of spice
my $HSPICE = ". /usr/local/packages/hspice/2003.03-SP1/bin/kshrc.meta\; /usr/local/packages/hspice/2003.03-SP1/bin/hspice ";

#my $synFile = $CellName . ".lib";
open(SYNOUT, ">$out") || die "Cannot open $out for write, $!\n";
print_lib_header($c, $d, $LibName);

foreach $FileName (@inFiles) {

my $CellName = $FileName;
$CellName =~ s/\..*//;
my $spiceFile = $CellName . ".spi";

my $arcStart = 0;
my %arcStuff;
my $arcNum = 0;
my %pins;
my $pinStart = 0;
my $pinNum = 0;

my $gap;
my %edgerates;
my %outloads;
my @rate;
my @load;
my $pinName;    
my $clock_pin;

# get pin caps
my %tmpHash = get_cap($spiceFile);

foreach (keys(%tmpHash)) {
  push(@{$pins{$_}->{$arcKeywords[23]}}, $tmpHash{$_});
}

# loop through the input file and get needed info
open(ARC, $FileName) || die "Cannot open input arc file, $FileName, $!\n";
while(<ARC>) {
  my @line = split;
  my $numWords = @line;
  # get pins
  if($line[0] =~ /^$arcKeywords[17]/i) {
      $pinStart = 1;
      $pinName = $line[1];
      next;
  }
  if($line[0] =~ /^$arcKeywords[18]/i) {
      $pinStart = 0;
      $pinNum++;
  }
  #default the clock_pi_ture to be false
  push(@{$pins{$pinName}->{$arcKeywords[25]}}, "false");
  
  # make load/edgerate arrays
  if($pinStart && $line[0]) {
    my $i = 0;
    # read in the input and output keyworded lines into the hash
    if($line[0] =~ /$arcKeywords[1]/ || $line[0] =~ /$arcKeywords[2]/) {
      do {
	   push(@{$pins{$pinName}->{$arcKeywords[21]}}, $line[$i++])
      } while($i < $numWords);
    } elsif($line[0] =~ /$arcKeywords[22]/) {
      # grab clk edge if pin will be used as a clock for setup
      my $sciClkRate = $line[1] . "e-12";
      push(@{ $pins{$pinName}->{$arcKeywords[22]} }, $sciClkRate);
    } elsif($line[0] =~ /$arcKeywords[24]/) {
          # grab the output function
          push(@{$pins{$pinName}->{$arcKeywords[24]}}, $line[1]);
    } elsif($line[0] =~ /$arcKeywords[25]/) {
          # Is this a clock pin?  If so, push onto the hash
          pop(@{$pins{$pinName}->{$arcKeywords[25]}});
          push(@{$pins{$pinName}->{$arcKeywords[25]}}, $line[1]);
    }
    # the arc is an input, get the start and end edge rate and make it
    # into an array of $c + 1 elements
    if ($pins{$pinName}->{$arcKeywords[21]}[0] =~ $arcKeywords[1]) {
      $gap = ($pins{$pinName}->{$arcKeywords[21]}[2] - $pins{$pinName}->{$arcKeywords[21]}[1]) / ($c - 1);
      my $f;
      for ($f = 0; $f < $c; $f++) {
        my $edge = ( $gap * $f) + $pins{$pinName}->{$arcKeywords[21]}[1];
        $edge =~ s/(\d+\.?\d?\d?\d?)\d*/\1/;
        my $sciedge = $edge . "e-12";
        push(@{$pins{$pinName}->{$arcKeywords[19]}},$sciedge);
      }
    } elsif($pins{$pinName}->{$arcKeywords[21]}[0] =~ $arcKeywords[2]) {
      # the pin is an output, get the start and end output loads and make in
      # into an array of $d + 1 elements
      $gap = ($pins{$pinName}->{$arcKeywords[21]}[2] - $pins{$pinName}->{$arcKeywords[21]}[1]) / ($d - 1);
      my $g;
      for ($g = 0; $g < $d; $g++) {
        my $load = ($gap * $g) + $pins{$pinName}->{$arcKeywords[21]}[1];
        $load =~ s/(\d+\.?\d?\d?\d?)\d*/\1/;
        my $sciload = $load . "e-15";
        push (@{$pins{$pinName}->{$arcKeywords[20]}},$sciload);
	 }
    }
  }
  # now get the arc
  # set flag if starting to go inside arc
  if($line[0] =~ /^$arcKeywords[8]/i) {
    $arcStart = 1;
    next;
  }
  # clear flag if at end of arc
  # and increment which arc we are on
  if($line[0] =~ /^$arcKeywords[9]/i) {
    $arcStart = 0;
    $arcNum++;
  }
  # loop on words on the line. push them into an array for that keyword ($line[0])
  if($arcStart && $line[0]) {
    my $i = 1;
    do {
      push(@{ $arcStuff{$arcNum}->{$line[0]} }, $line[$i++]);
    } while($i < $numWords);
  }
}
# get pin order from subckt call
my @ordered_pins = getPinOrder($spiceFile);

# store default file handle for printing
$arcNum = 0;
# print tests and measures
my $ofh = select;
foreach $arcNum (keys(%arcStuff)) {
  #create and open unique filename for each arc
  my $spiceOut = $CellName . ".arc" . $arcNum . ".ctrl";
  open(SPICEOUT, ">$spiceOut") || die "Cannot open output file, $spiceOut, $!\n";
  # auto-alias default filehandle for printing to the current spice output file
  select(SPICEOUT);
  # print default headers
  printSpiceHeaders($arcNum, *SPICEOUT, $spiceFile);
  # just in case the spice output looks strange
  # run in debug mode to print out the arcs
  # to ensure they are being read in correctly
  if($debug) {
    my $keyword = "";
    my $createdLine = "$arcKeywords[8] $arcNum\n";
    foreach $keyword (keys(%{ $arcStuff{$arcNum} } )) {
      my $word = "";
      $createdLine .= $keyword;
      foreach $word (@{ $arcStuff{$arcNum}->{$keyword} }) {
	$createdLine .= " " . $word;
      }
      $createdLine .= "\n";
      $createdLine .= "$arcKeywords[9]\n\n";
      print $createdLine;
    }         
  }           # END if($debug)
  # print subckt instantiation
  my $subci = "$spiceKeywords[0] ";
  foreach (@ordered_pins) {
    $subci .= "$_ ";
  }
  $subci .= "$CellName\n";
  print "$subci\n";
  # get edge rates array
  my $inputPinName = $arcStuff{$arcNum}->{$arcKeywords[1]}[0];
  my $outputPinName = $arcStuff{$arcNum}->{$arcKeywords[2]}[0];
  my $clkPinName = $arcStuff{$arcNum}->{$arcKeywords[3]}[0];
  my $clkEdgeRate = $pins{$clkPinName}->{$arcKeywords[22]}[0];
  my $sizeOfEdgeArray = $#{$pins{$inputPinName}->{$arcKeywords[19]}};
  my $startEdgeRate = $pins{$inputPinName}->{$arcKeywords[19]}[0];
  my $endEdgeRate = $pins{$inputPinName}->{$arcKeywords[19]}[$sizeOfEdgeArray];
  my $incrEdgeRate = "\'\($endEdgeRate - $startEdgeRate\) \/ \($sizeOfEdgeArray\)\'";
  # print spice stuff and measures
  # get input and clock switch directions
  my @inV;
  my @clockV;
  # determine input switch direction
  if($arcStuff{$arcNum}->{$arcKeywords[1]}[1] =~ /$arcKeywords[13]/) {
    @inV = ( "0", $spiceKeywords[1] );
  } elsif($arcStuff{$arcNum}->{$arcKeywords[1]}[1] =~ /$arcKeywords[14]/) {
    @inV = ( $spiceKeywords[1], "0" );
  }
  # determine clock switch direction
  if($arcStuff{$arcNum}->{$arcKeywords[3]}[1] =~ /$arcKeywords[13]/) {
    @clockV = ( "0", $spiceKeywords[1] );
  } elsif($arcStuff{$arcNum}->{$arcKeywords[3]}[1] =~ /$arcKeywords[14]/) {
    @clockV = ( $spiceKeywords[1], "0" );
  }
  # print params
  print ".param $spiceKeywords[4] = $startEdgeRate\n";
  if($clkEdgeRate) {
    print ".param $spiceKeywords[7] = $clkEdgeRate\n";
  }
  print ".param $spiceKeywords[3] = $pins{$outputPinName}->{$arcKeywords[20]}[0]\n";
  # print sources for pins
  if($arcStuff{$arcNum}->{$arcKeywords[0]}[0] =~ /$arcKeywords[10]/) {
    # if the arc is a delay, make a pwl for the input keyword
    print "v$arcStuff{$arcNum}->{$arcKeywords[1]}[0] $arcStuff{$arcNum}->{$arcKeywords[1]}[0] 0 pwl( 0 $inV[0] 100p $inV[0] \'100e-12 + $spiceKeywords[4]\' $inV[1] \'(100e-12 + $spiceKeywords[4]) * 2\' $inV[1])\n";
  } elsif($arcStuff{$arcNum}->{$arcKeywords[0]}[0] =~ /$arcKeywords[11]/) {
    # arc is setup, do spice bisection optimization to find min setup
    print ".param $spiceKeywords[5] = \'400e-12\'\n";
    my $tmpSetupEdge = $spiceKeywords[6] . "CLK_". $arcStuff{$arcNum}->{$arcKeywords[3]}[1] . "_DATA_" . $arcStuff{$arcNum}->{$arcKeywords[1]}[1];
    print ".param $tmpSetupEdge = opt1 (0p, \'$spiceKeywords[5]\', \'-1 * $spiceKeywords[5]\')\n";
    # make pwl for both clock and input keywords
    print "v$arcStuff{$arcNum}->{$arcKeywords[1]}[0] $arcStuff{$arcNum}->{$arcKeywords[1]}[0] 0 pwl( 0 $inV[0] \'$spiceKeywords[5] - $tmpSetupEdge\' $inV[0] \'$spiceKeywords[5] - $tmpSetupEdge + $spiceKeywords[4]\' $inV[1] \'$spiceKeywords[5] * 10\' $inV[1])\n";
    print "v$arcStuff{$arcNum}->{$arcKeywords[3]}[0] $arcStuff{$arcNum}->{$arcKeywords[3]}[0] 0 pwl( 0 $clockV[0] \'$spiceKeywords[5]\' $clockV[0] \'$spiceKeywords[5] + $spiceKeywords[7]\' $clockV[1] \'$spiceKeywords[5] * 10\' $clockV[1])\n";
  }
  # make dc sources for all tie keyworded inputs
  for(my $i=0; $i<$#{ $arcStuff{$arcNum}->{$arcKeywords[4]} }; $i+=2) {
    my $pin   = @{ $arcStuff{$arcNum}->{$arcKeywords[4]}}[$i];
    my $value = @{ $arcStuff{$arcNum}->{$arcKeywords[4]}}[$i+1];
    my $tieV;
    if($value =~ /$arcKeywords[15]/) {
      $tieV = $spiceKeywords[1];
    } elsif($value =~ /$arcKeywords[16]/) {
      $tieV = "0";
    }
    print "v$pin $pin 0 $tieV\n";
  }
  # make ic statements
  for(my $i=0; $i<$#{ $arcStuff{$arcNum}->{$arcKeywords[6]} }; $i+=2) {
    my $pin   = $arcStuff{$arcNum}->{$arcKeywords[6]}[$i];
    my $value = $arcStuff{$arcNum}->{$arcKeywords[6]}[$i+1];
    my $icV;
    if($value =~ /$arcKeywords[15]/) {
      $icV = $spiceKeywords[1];
    } elsif($value =~ /$arcKeywords[16]/) {
      $icV = "0";
    }
    print ".ic $spiceKeywords[0].$pin $icV\n";
  }
  # make PWL(VCVS?) for all drive keyworded inputs
  for(my $i=0; $i<$#{ $arcStuff{$arcNum}->{$arcKeywords[5]} }; $i+=2) {
    my $pin   = $arcStuff{$arcNum}->{$arcKeywords[5]}[$i];
    my $value = $arcStuff{$arcNum}->{$arcKeywords[5]}[$i+1];
    my @driveV;
    if($value =~ /$arcKeywords[13]/) {
      @driveV = ( "0", $spiceKeywords[1] );
    } elsif($value =~ /$arcKeywords[14]/) {
      @driveV = ( $spiceKeywords[1], "0" );
    }
    print "v$pin $pin 0 pwl ( 0 $driveV[0] 100p $driveV[0] \'100e-12 + $spiceKeywords[4]\' $driveV[1] \'(100e-12 + $spiceKeywords[4]) * 2\' $driveV[1])\n";
  }
  # add load cap to output pin
  print "cload $arcStuff{$arcNum}->{$arcKeywords[2]}[0] 0 $spiceKeywords[3]\n";
  if(@{ $arcStuff{$arcNum}->{$arcKeywords[0]} }[0] =~ /$arcKeywords[10]/) {
    # arc is delay; print tran and appropriate measures, add load alters
    # add tran statement
    print ".tran 1p 1n\n"; 
    # Add delay and output edge rate measures
    print ".meas tran cell_$arcStuff{$arcNum}->{$arcKeywords[1]}[1] trig v($arcStuff{$arcNum}->{$arcKeywords[1]}[0]) val = \'0.5 * $spiceKeywords[1]\' $arcStuff{$arcNum}->{$arcKeywords[1]}[1] = 1 targ v($arcStuff{$arcNum}->{$arcKeywords[2]}[0]) val = \'0.5 * $spiceKeywords[1]\' $arcStuff{$arcNum}->{$arcKeywords[2]}[1] = 1\n";
    my $tmpOutEdge = $arcStuff{$arcNum}->{$arcKeywords[2]}[1] . "_tran";
    my @measDir;
    if($arcStuff{$arcNum}->{$arcKeywords[2]}[1] =~ /$arcKeywords[13]/) {
          $measDir[0] = "0.2";
          $measDir[1] = "0.8";
    } else {
          $measDir[0] = "0.8";
          $measDir[1] = "0.2";
    }
    print ".meas tran $tmpOutEdge trig v($arcStuff{$arcNum}->{$arcKeywords[2]}[0]) val = \'$measDir[0] * $spiceKeywords[1]\' $arcStuff{$arcNum}->{$arcKeywords[2]}[1] = 1 targ v($arcStuff{$arcNum}->{$arcKeywords[2]}[0]) val = \'$measDir[1] * $spiceKeywords[1]\' $arcStuff{$arcNum}->{$arcKeywords[2]}[1] = 1\n";  
  } elsif(@{ $arcStuff{$arcNum}->{$arcKeywords[0]} }[0] =~ /$arcKeywords[11]/) {
    # arc is setup; finish setting up bisection, add necessary measures
    # then add load and edge rate alters
    print ".MODEL OptMod opt Method=bisection\n";
    my $state_dir;
    if($arcStuff{$arcNum}->{$arcKeywords[7]}[1] =~ /$arcKeywords[13]/) {
      $state_dir = "max";
    } elsif($arcStuff{$arcNum}->{$arcKeywords[7]}[1] =~ /$arcKeywords[14]/) {
      $state_dir = "min";
    }
    print ".meas tran $spiceKeywords[8] $state_dir v($spiceKeywords[0].$arcStuff{$arcNum}->{$arcKeywords[7]}[0]) goal = \'$arcStuff{$arcNum}->{$arcKeywords[7]}[2] * $spiceKeywords[1]\'\n";
#    print ".tran 1p \'$spiceKeywords[5] * 2\' sweep Optimize = opt1 Result = $spiceKeywords[8] Model = OptMod\n";
    print ".tran 1p 1n sweep Optimize = opt1 Result = $spiceKeywords[8] Model = OptMod\n";
  }
# print alters for loads and edges
  foreach my $load (@{ $pins{$outputPinName}->{$arcKeywords[20]} }) {
    foreach my $rate (@{ $pins{$inputPinName}->{$arcKeywords[19]} }) {
      my $alter_title = "load " . $load . " edge " . $rate . " ";
      print ".alter\n";
      print ".title $alter_title\n";
      print ".param $spiceKeywords[3] = $load\n";
      print ".param $spiceKeywords[4] = $rate\n"
    }
  }
  print ".end";
  close(SPICEOUT);
} # END foreach $arcNum
  
select($ofh);

$arcNum = 0;
my %outPinHash;
print SYNOUT "\ncell($CellName){\n";
foreach $arcNum (keys(%arcStuff)) {
  my $arcName = $CellName . ".arc$arcNum";
  my $arcFile = $arcName . ".ctrl";
  my $outFile = $arcFile . ".out";
  # run spice for current arc
  my $spiceOut = `$HSPICE $arcFile > $outFile`;
  # collect any errors
  if ($spiceOut =~ /error/ || $spiceOut =~ /aborted/) {
    print "ERROR: hspice aborted\n\tCheck $outFile for errors\n\n";
  } else {
    # if no errors, grab measure results from .mt* files
    # and print them formated into the syn file
    if($arcStuff{$arcNum}->{$arcKeywords[0]}[0] =~ /setup/) {
      print_input_timing($arcStuff{$arcNum}->{$arcKeywords[2]}[0], $pins{$arcStuff{$arcNum}->{$arcKeywords[1]}[0]}->{$arcKeywords[25]}[0],0, $CellName, $arcNum, $arcStuff{$arcNum}->{$arcKeywords[1]}[0], $arcStuff{$arcNum}->{$arcKeywords[3]}[1], $pins{$arcStuff{$arcNum}->{$arcKeywords[1]}[0]}->{$arcKeywords[23]}[0], $arcStuff{$arcNum}->{$arcKeywords[3]}[0], $arcStuff{$arcNum}->{$arcKeywords[1]}[1], $d, $c, @{ $pins{$arcStuff{$arcNum}->{$arcKeywords[2]}[0]}->{$arcKeywords[20]} }, @{ $pins{$arcStuff{$arcNum}->{$arcKeywords[1]}[0]}->{$arcKeywords[19]} });
    } else {
      # need to do more complicated mess if it is a delay arc
      # must first combine all delays together and put them all
      # into the output pin specified in the arc file, all at once.
      my $tmpName = $CellName . ".arc" . $arcNum;
      my %tmpHash = read_mt0($tmpName);
      my $outPinName = $arcStuff{$arcNum}->{$arcKeywords[2]}[0];
      my $relPinName = $arcStuff{$arcNum}->{$arcKeywords[1]}[0];
      my $timingType = $arcStuff{$arcNum}->{$arcKeywords[1]}[1];
      my $cellDir = $arcStuff{$arcNum}->{$arcKeywords[1]}[1];
      foreach my $load (@{ $pins{$outPinName}->{$arcKeywords[20]} }) {
          foreach my $edge (@{ $pins{$relPinName}->{$arcKeywords[19]} }) {
               @{ $outPinHash{$outPinName}->{$relPinName}->{$timingType}->{$cellDir}->{'edges'}->{$edge}->{$load} } = @{ $tmpHash{'edges'}->{$load}->{$edge} };
               @{ $outPinHash{$outPinName}->{$relPinName}->{$timingType}->{$cellDir}->{'timing'}->{$edge}->{$load} } = @{ $tmpHash{'timing'}->{$load}->{$edge} };
          }
       }
    }
    # cleanup spice junk
    if($debug == 0) {
      my $arcMess = $CellName . ".arc" . $arcNum;
      clean_file($arcMess);
    }
  }
}
# make sure we did not leave off any input pin which
# does not have an input setup associated with it
print_rest_timing(\%arcStuff, \%pins);
# now go back and print the output pin data
print_output_timing($d, $c, $pins{$arcStuff{'0'}->{$arcKeywords[2]}[0]}->{$arcKeywords[24]}[0], \%outPinHash, \%pins, @{ $pins{$arcStuff{'0'}->{$arcKeywords[2]}[0]}->{$arcKeywords[20]} }, @{ $pins{$arcStuff{'0'}->{$arcKeywords[1]}[0]}->{$arcKeywords[19]} });
print SYNOUT "\t}\n}";
}

# close out the .lib file
#print_lib_footer();
print SYNOUT "\n}";
close(SYNOUT);

# .include \"./mos013.txt\"\
# Print default Headers. Need to fix filehandle passing from calling loop to this subroutine
sub printSpiceHeaders {
  my $arcNum = shift;
  my $fileHandle = shift;
  my $incFile = shift;
  my $whoAmI = `whoami`;
  my $today = `date`;
  chomp($whoAmI);
  chomp($today);
  print $fileHandle ("\* $incFile:arc$arcNum created $today by $whoAmI\
.title $incFile:arc$arcNum\
.options SPICE NOMOD AUTOSTOP ACCT POST=0 CO=132\
.param ttra=591.8p LV=0.2p\
.temp 27\
.include \"./mos009_FF.txt\"\
.global vdd\
.global gnd\
.param $spiceKeywords[1] = 1.5\
vvdd vdd 0 $spiceKeywords[1]\
vgnd gnd 0 0\
.include \"./$incFile\"\
\n");
}

sub getPinOrder {
  my $file = shift;
  my @ordered;
  my $cellName = $file;
  $cellName =~ s/\.sp.*//;
  open(FILE, $file) || die "Cannot open $file for read inside main::getPinOrder, $!\n";
  while(<FILE>) {
    if(/\.subckt/) {
      my $pin;
      foreach $pin (split) {
        if($pin !~ /\.subckt/ && $pin !~ /$cellName/) {
          push(@ordered, $pin);
        }
      }
      close(FILE);
      return @ordered;
    }
  }
}

sub get_cap {
  my %pin_hash;
  my @pin_array;
  my $spice_file = shift;
  my $subckt = $spice_file;
  $subckt =~ s/\..*//;

  @pin_array = getPinOrder($spice_file);
  open (CONTROL, ">${subckt}.ctrl") || die "Could not open ${subckt}.ctrl for write";

  printSpiceHeaders("Cap Sim", *CONTROL, $spice_file);
  
  print CONTROL "X${subckt} @pin_array $subckt\n";
  print CONTROL ".tran 0.1p 400p\n";

  #Need to do one simulation for each pin.  Do sim on output and through away later.
  my $count = 0;
  foreach my $i (@pin_array) {
    #We are integrating the current and dividing by the voltage to get the cap
    print CONTROL ".measure tran charge integ(I(V${i})) from=10p to=300p\n";
    print CONTROL ".measure tran cap param = '-1*charge/vp'\n";
    foreach my $j (@pin_array) {
#      if($j =~ $i){print CONTROL "V${j} $j 0 pwl(0n 0v 10p 0v 300p vp 400p vp)\n";}
#      else {print CONTROL "V${j} $j 0 vp\n";}
      print CONTROL "V${j} $j 0 pwl(0n 0v 10p 0v 300p vp 400p vp)\n";
    }
    if ($count < scalar(@pin_array)-1){print CONTROL ".alter\n";}
    $count++;
  }
	
  print CONTROL ".end\n";
  close CONTROL;
    
  #Run spice sim and check for errors
  my $spiceOut = $spice_file . ".out";
  my $error = `$HSPICE ${subckt}.ctrl > $spiceOut`;
  if ($error =~ /error/ || $error =~ /aborted/) {
    print STDERR "Hspice Simulation failed. Please check setup and rerun\n";
  }
  
  $count = 0;
  #get pin data into a hash 
  foreach my $i (@pin_array) {
    open (OUTDATA, "${subckt}.mt${count}");
    while (<OUTDATA>) {
      my @line = split();
      #This is kind of a kluge, but it works.  If it breaks later we will fix it
      if($line[1] =~ /^\d/){
	$pin_hash{$i} = ($line[1] * $capMult);
      }
    }
    $count++;
    close OUTDATA;
  }
  clean_file(${subckt});
  return %pin_hash   
}

sub read_mt0
{
        my $circuit_arc = @_[0];

        #We need to loop through all the mt files for a given arc
        my $meas_count = `ls -ltr ${circuit_arc}.mt* | wc -l`;
        #Correct for extra mt0
        $meas_count--;

        my %return_hash;

        my $item2 = "timing";
        my $item3 = "edges";
        my $ext = 1;

        for (my $i=1; $i<=$meas_count; $i++)
        {
                my $edge;
                my $load;
                my $timing;
                my $timing_edge;
                my $j_count = 0;
                my $k_count = 0;

                my $place = 0;

                #mt files have letter extensions is greater than 9.  Not sure what happens if we go over z?
                if ($ext =~/\d+/ && $ext > 9 && $ext < 36){$ext = "a";}
                my $file = $circuit_arc . ".mt" . $ext;
                open (MT0, "<$file") || die "Could not open $file for read";
                while (<MT0>)
                {
                        if($_ =~ /^\$/){next;}
                        if($_ =~ /index/){$place=1;}
                        my @line = split;

                        if($line[0] =~ /\.TITLE/)
			  {
			    $load=$line[3];
			    $edge=$line[5];
			  }
                        $edge =~ s/\'//;
			if($line[0] =~ /\d+/)
			  {
			    $timing = ($line[0+${place}] * $timingMult);
			    $timing_edge = ($line[1+${place}] * $timingMult);
		            last;
	                  }
                }

                #You will need to sort these when you are done
                push (@{$return_hash{$item2}->{$load}->{$edge}}, $timing);
                push (@{$return_hash{$item3}->{$load}->{$edge}}, $timing_edge);

                close MT0;
                if($ext =~ /z/) {
                    $ext = 35;
               }
                $ext++;
        }
        return %return_hash
}

sub print_lib_header {
     my $edge = shift;
     my $load = shift;
     my $libName = shift;
     my @edges;
     my @loads;
     for(my $i=1; $i<=$edge; $i++) {
          push(@edges, $i);
     }
     for(my $j=1; $j<=$load; $j++) {
          push(@loads, $j);
     }
     my $edgeNums = join(", ", @edges);
     my $loadNums = join(", ", @loads);
     
  print SYNOUT "library ($libName) {\n";
  print SYNOUT "\tdelay_model : table_lookup\;\n";
  
  print SYNOUT "\tlibrary_features (report_delay_calculation) \;\n";
  
  print SYNOUT "\tdefine(\"is_gated_clock\", \"cell\", \"string\")\;\n";
  print SYNOUT "\tin_place_swap_mode : match_footprint;\n\n";

  print SYNOUT "\tlu_table_template (\"input${edge}x${load}\") {\n";
  print SYNOUT "\t\tvariable_1 : input_net_transition\;\n";
  print SYNOUT "\t\tindex_1 (\"$edgeNums\")\;\n";
  print SYNOUT "\t\tvariable_2 : total_output_net_capacitance\;\n";
  print SYNOUT "\t\tindex_2 (\"$loadNums\")\;\n";
  print SYNOUT "\t}\n\n";

  print SYNOUT "\tlu_table_template (\"setup${edge}x${load}\") {\n";
  print SYNOUT "\t\tvariable_1 : related_pin_transition\;\n";
  print SYNOUT "\t\tindex_1 (\"$edgeNums\")\;\n";
  print SYNOUT "\t\tvariable_2 : related_out_total_output_net_capacitance\;\n";
  print SYNOUT "\t\tindex_2 (\"$loadNums\")\;\n";
  print SYNOUT "\t}\n\n";

  print SYNOUT "\tlu_table_template (\"output${edge}x${load}\") {\n";
  print SYNOUT "\t\tvariable_1 : input_net_transition\;\n";
  print SYNOUT "\t\tindex_1 (\"$edgeNums\")\;\n";
  print SYNOUT "\t\tvariable_2 : total_output_net_capacitance\;\n";
  print SYNOUT "\t\tindex_2 (\"$loadNums\")\;\n";
  print SYNOUT "\t}\n\n";
  
  
  print SYNOUT "\ttime_unit : \"1ps\"\;\n";
  print SYNOUT "\tvoltage_unit : \"1V\"\;\n";
  print SYNOUT "\tcurrent_unit : \"1mA\"\;\n";
  print SYNOUT "\tpulling_resistance_unit : \"1kohm\"\;\n";
  print SYNOUT "\tcapacitive_load_unit(1,ff)\;\n\n";
  
  print SYNOUT "\tdefault_inout_pin_cap : 0.0\;\n";
  print SYNOUT "\tdefault_input_pin_cap : 0.0\;\n";
  print SYNOUT "\tdefault_output_pin_cap : 0.0\;\n";
  print SYNOUT "\tdefault_fanout_load    : 0.0\;\n";
  print SYNOUT "\tinput_threshold_pct_fall        : 50.0\;\n";
  print SYNOUT "\tinput_threshold_pct_rise        : 50.0\;\n";
  print SYNOUT "\toutput_threshold_pct_fall       : 50.0\;\n";
  print SYNOUT "\toutput_threshold_pct_rise       : 50.0\;\n";
  print SYNOUT "\tslew_lower_threshold_pct_rise   : 20.0\;\n";
  print SYNOUT "\tslew_upper_threshold_pct_rise   : 80.0\;\n";
  print SYNOUT "\tslew_lower_threshold_pct_fall   : 20.0\;\n";
  print SYNOUT "\tslew_upper_threshold_pct_fall   : 80.0\;\n\n";
  
  print SYNOUT "\tdefault_wire_load_capacitance : 0.1\;\n";
  print SYNOUT "\tdefault_wire_load_resistance : 1.0e-3\;\n";
  print SYNOUT "\tdefault_wire_load_area       : 0.0\;\n\n";
  
  print SYNOUT "\tnom_process                  :  1.0\;\n";
  print SYNOUT "\tnom_temperature              : 27.0\;\n\n";
  
  print SYNOUT "\twire_load(m1_m2_avg) {\n";
  print SYNOUT "\tresistance : 0.26e-3 \;\n";
  print SYNOUT "\tcapacitance : 0.26 \;\n";
  print SYNOUT "\tarea : 0.0 \;\n";
  print SYNOUT "\tslope : 0.3 \;\n";
  print SYNOUT "\tfanout_length(1,7) \;\n";
  print SYNOUT "\t}\n\n";
}
    
sub print_input_timing
{
        my $output_pin = shift;
        my $clock_pin_true = shift;
        my $noSetup = shift;
        my $CellNameCarry = $CellName;
        my $CellName = shift;
        my $arcNum = shift;
        my $pin = shift;
        my $timing_type = shift;
        my $pin_cap = shift;
        my $related = shift;
        my $relatedDir = shift;
        my $arcFile = $CellName . ".arc" . $arcNum;
        my %return_hash = read_mt0($arcFile);
        my $item2 = "timing";
        my $item3 = "edges";
        my $loadnum = shift;
        my $edgenum = shift;
        my @load;
        my @in_edge;
        for(my $i=0; $i<$loadnum; $i++) {
		$load[$i] = shift;
        }
        for(my $i=0; $i<$edgenum; $i++) {
		$in_edge[$i] = shift;
        }

       if ($timing_type =~ "fall")
	  {
	    $timing_type = "setup_falling";
	  } elsif($timing_type =~ "rise") {
	    $timing_type = "setup_rising";
	  }
       my @nonSciEdges = @in_edge;
       my @nonSciLoads = @load;
        for(my $i=0; $i<$loadnum; $i++) {
		$nonSciLoads[$i] *= $capMult;
        }
        for(my $i=0; $i<$edgenum; $i++) {
		$nonSciEdges[$i] *= $timingMult;
        }

        print SYNOUT "\tpin (${pin}) {\n";
        print SYNOUT "\t\tdirection : input \;\n";
        print SYNOUT "\t\tcapacitance : ${pin_cap}\;\n";
		print SYNOUT "\t<DATA>, Cell Name:, ${CellNameCarry}, Pin Name:, ${pin}, input cap:, ${pin_cap} \n";
        #If a clock pin is defined, we need to tell the compiler
        if($clock_pin_true =~ /true/){
        print SYNOUT "\t\tclock : true\;\n";
        }
        
        if($noSetup == 0) {
             print SYNOUT "\t\ttiming () {\n";
             print SYNOUT "\t\ttiming_type : ${timing_type}\;\n";

             if ($relatedDir =~ "fall")
             {
                     print SYNOUT "\t\t\tfall_constraint (setup${c}x${d}) {\n";
             }
             if ($relatedDir =~ "rise")
             {
                     print SYNOUT "\t\t\trise_constraint (setup${c}x${d}) {\n";
             }

             my $index1 = "\t\t\t\tindex_1 ( \"";
             $index1 .= join(", ", @nonSciEdges);
             $index1 .= "\" )\;\n";
             print SYNOUT $index1;
             my $index2 = "\t\t\t\tindex_2 ( \"";
             $index2 .= join(", ", @nonSciLoads);
             $index2 .= "\" ) \;\n";
             print SYNOUT $index2;

             print SYNOUT "\t\t\t\tvalues ( \\\n";
              my $loopNum = 0;
             foreach my $val2 (@in_edge)
             {
                     my $tmp = "";
                     my @table;
                     foreach my $val1 (@load)
                     {
                             foreach my $val3 (@{$return_hash{$item2}->{$val1}->{$val2}})
                             {
                                     push (@table, $val3);
                             }
                     }
                     $tmp = join(", ",@table);
                     if($loopNum < ($edgenum - 1)) {
                         print SYNOUT "\t\t\t\t\t\"$tmp\",\\";
                     } else {
                         print SYNOUT "\t\t\t\t\t\"$tmp\"\\";
                     }
                     print SYNOUT "\n";
                     $loopNum++;

             }

             print SYNOUT "\t\t\t\t) \; \n";
             print SYNOUT "\t\t\t}\n";


             print SYNOUT "\t\t\trelated_pin : \"${related}\"\;\n";
             print SYNOUT "\t\t\trelated_output_pin : \"${output_pin}\"\;\n";
             print SYNOUT "\t\t}\n";
        }
        print SYNOUT "\t}\n";
}

sub clean_file
  {
     my $input = @_[0];
 
     my $file0 = $input.".mt*";
     my $file1 = $input.".pa*";
     my $file2 = $input.".ic*";
     my $file3 = $input.".st*";
     my $file4 = $input.".tr*";
 
     `rm $file0`;
     `rm $file1`;
     `rm $file2`;
     `rm $file3`;
     `rm $file4`;
     return;
  }

sub print_output_timing {
	my $loadnum = shift;
	my $edgenum = shift;
     my $function = shift;
	my $hashRef = shift;
     my $pinHashRef = shift;
     my @loads;
     my @in_edge;
     my %tmpHash;

	for(my $i=0; $i<$loadnum; $i++) {
          $loads[$i] = shift;
     }
     for(my $i=0; $i<$edgenum; $i++) {
          $in_edge[$i] = shift;
     }
       my @nonSciEdges = @in_edge;
       my @nonSciLoads = @loads;
        for(my $i=0; $i<$loadnum; $i++) {
		$nonSciLoads[$i] *= $capMult;
        }
        for(my $i=0; $i<$edgenum; $i++) {
		$nonSciEdges[$i] *= $timingMult;
        }

     foreach my $pinName (keys(%$pinHashRef)) {
          if($$pinHashRef{$pinName}->{'direction'}[0] =~ /input/) {
               $tmpHash{$pinName} = 1;
          }
     }
     
PINNAME:     foreach my $pinName (keys(%$pinHashRef)) {
          if($$pinHashRef{$pinName}->{'direction'}[0] =~ /input/) {
               if(!defined($$pinHashRef{$pinName}->{'clk_edge_rate'}[0])) {
                    delete($tmpHash{$pinName});
               }
          }
     }
     my $Combo = 1;
     foreach my $foo (keys(%tmpHash)) {
          $Combo = 0;
     }

	
#print DumperX(%$hashRef);

	foreach my $pinName (keys(%$hashRef)) {
          print SYNOUT "\tpin (${pinName}) {\n";
          print SYNOUT "\t\tdirection : output \;\n";
          print SYNOUT "\t\tfunction : \"$function\" \;\n";
          foreach my $relPin (keys(%{$$hashRef{$pinName}})) {
               #remove duplicate print statements
               my $found = 0;
               my $both =0;
               print SYNOUT "\t\ttiming() {\n";
               print SYNOUT "\t\t\trelated_pin : \"$relPin\" \;\n";
               
               #We need to figure out if both directions are switching.
               foreach my $timingType (keys(%{$$hashRef{$pinName}->{$relPin}})) {$both ++;}
               
               foreach my $timingType (keys(%{$$hashRef{$pinName}->{$relPin}})) {
                    my $timing_type;
                    if($timingType =~ /rise/) {
                         $timing_type = "combinational_rise";  
                    } else {
                         $timing_type = "combinational_fall";
                    }
                    
                    if(! $Combo  && !$found && ($both < 2)) {
                         print SYNOUT "\t\t\ttiming_type : $timing_type \;\n";
                         print SYNOUT "\t\t\ttiming_sense : positive_unate \;\n";
                         $found = 1; 
                    }
                    if(! $Combo  && ($both == 2)) {
                         print SYNOUT "\t\t\ttiming_type : combinational \;\n";
                         print SYNOUT "\t\t\ttiming_sense : positive_unate \;\n";
                         $found = 1;
                         $both = 0;

                    }
                    
                    foreach my $cellDir (keys(%{$$hashRef{$pinName}->{$relPin}->{$timingType}})) {
                         my $cell_dir;
						 my $cell_desc;
                         if($cellDir =~ /rise/) {
						 	  $cell_desc = "cell_rise";	# MY EDIT 26Aug05
                              $cell_dir = "cell_rise(output" . $edgenum . "x" . $loadnum . "){";
                         } else {
						 	  $cell_desc = "cell_fall";	# MY EDIT 26Aug05
                              $cell_dir = "cell_fall(output" . $edgenum . "x" . $loadnum . "){";
                         }
                         print SYNOUT "\t\t\t$cell_dir\n";
                         
                         my $index1 = "\t\t\t\tindex_1 ( \"";
                         $index1 .= join(", ", @nonSciEdges);
                         $index1 .= "\" )\;\n";
#                         my $index2 = "\t\t\t\tindex_2 ( \"";
 #                        $index2 .= join(", ", @nonSciLoads);
  #                       $index2 .= "\" ) \;\n";
                         print SYNOUT $index1;
                         my $index2 = "<DATA>,";
						 $index2 .= join(" ", $CellName, ", ", $relPin, ", CAP, ,");      # MY EDIT 26Aug05
						 $index2 .= join(", ,  ,  ");      # MY EDIT 26Aug05
                         $index2 .= join(", ", @nonSciLoads);
                         $index2 .= "\n";
                         print SYNOUT $index2;    
                         
                         print SYNOUT "\t\t\t\tvalues ( \\\n";
                         my $loopNum = 0;
                         foreach my $edge (@in_edge) {
                              my @line;
							  print SYNOUT "<DATA>, ";      # MY EDIT 26Aug05
							  print SYNOUT $CellName;      # MY EDIT 26Aug05
							  print SYNOUT ", ";      # MY EDIT 26Aug05
							  print SYNOUT $relPin;      # MY EDIT 26Aug05
							  print SYNOUT ", ";      # MY EDIT 26Aug05
							  print SYNOUT $cell_desc;      # MY EDIT 26Aug05
							  print SYNOUT ", ";      # MY EDIT 26Aug05
							  print SYNOUT $edge;      # MY EDIT 26Aug05
                              my $tmp;
                              foreach my $load (@loads) {
                                   foreach my $timing (@{ $$hashRef{$pinName}->{$relPin}->{$timingType}->{$cellDir}->{'timing'}->{$edge}->{$load}}) {
                                        push(@line, $timing);
                                   }
                              }
                              $tmp = join(", ",@line);
                               if($loopNum < ($edgenum - 1)) {
                                   #print SYNOUT "\t\t\t\t\t\"$tmp\",\\";  # MY EDIT 26Aug05
                                   print SYNOUT ", $tmp";   # MY EDIT 26Aug05
                               } else {
                                   # print SYNOUT "\t\t\t\t\t\"$tmp\"\\";  # MY EDIT 26Aug05
                                    print SYNOUT ", $tmp ";  # MY EDIT 26Aug05
                              }
                              print SYNOUT "\n";
                              $loopNum++;
                         }
                         print SYNOUT "\t\t\t\t) \; \n";
                         print SYNOUT "\t\t\t}\n";
                         my $cell_tran;
                         if($cellDir =~ /rise/) {
						 	  $cell_desc = "rise_transition";	# MY EDIT 26Aug05
                              $cell_tran = "rise_transition(output" . $edgenum . "x" . $loadnum . "){";
                         } else {
						 	  $cell_desc = "fall_transition";	# MY EDIT 26Aug05
                              $cell_tran = "fall_transition(output" . $edgenum . "x" . $loadnum . "){";
                         }
                         print SYNOUT "\t\t\t$cell_tran\n";
                         
                         my $index1 = "\t\t\t\tindex_1 ( \"";
                         $index1 .= join(", ", @nonSciEdges);
                         $index1 .= "\" )\;\n";
                         print SYNOUT $index1;
                         my $index2 = "<DATA>,";
						 $index2 .= join(" ", $CellName, ", ", $relPin, ", CAP, ,");      # MY EDIT 26Aug05
						 $index2 .= join(", ,  ,  ");      # MY EDIT 26Aug05
                         $index2 .= join(", ", @nonSciLoads);
                         $index2 .= "\n";
#                         my $index2 = "\t\t\t\tindex_2 ( \"";
#                         $index2 .= join(", ", @nonSciLoads);
#                         $index2 .= "\" ) \;\n";
                         print SYNOUT $index2;    
                         
                         print SYNOUT "\t\t\t\tvalues ( \\\n";
                         my $loopNum = 0;
                         foreach my $edge (@in_edge) {
                              my @line;
							  print SYNOUT "<DATA>, ";      # MY EDIT 26Aug05
							  print SYNOUT $CellName;      # MY EDIT 26Aug05
							  print SYNOUT ", ";      # MY EDIT 26Aug05
							  print SYNOUT $relPin;      # MY EDIT 26Aug05
							  print SYNOUT ", ";      # MY EDIT 26Aug05
							  print SYNOUT $cell_desc;      # MY EDIT 26Aug05
							  print SYNOUT ", ";      # MY EDIT 26Aug05
							  print SYNOUT $edge;      # MY EDIT 26Aug05
                              my $tmp;
                              foreach my $load (@loads) {
                                   foreach my $timing (@{ $$hashRef{$pinName}->{$relPin}->{$timingType}->{$cellDir}->{'edges'}->{$edge}->{$load}}) {
                                        push(@line, $timing);
                                   }
                              }
                              $tmp = join(", ",@line);
                               if($loopNum < ($edgenum - 1)) {
                                  # print SYNOUT "\t\t\t\t\t\"$tmp\",\\";
                                   print SYNOUT ", $tmp";   # MY EDIT 26Aug05
                               } else {
                                  # print SYNOUT "\t\t\t\t\t\"$tmp\"\\";
                                   print SYNOUT ", $tmp";   # MY EDIT 26Aug05
                               }
                   
                              print SYNOUT "\n";
                              $loopNum++;
                         }
                         print SYNOUT "\t\t\t\t) \; \n";
                         print SYNOUT "\t\t\t}\n";
                    }
               }
               print SYNOUT "\t\t}\n";
          }
     }
}

sub print_rest_timing {
     my $arcHashRef = shift;
     my $pinHashRef = shift;
     my %tmpHash;
     foreach my $pinName (keys(%$pinHashRef)) {
          if($$pinHashRef{$pinName}->{'direction'}[0] =~ /input/) {
               $tmpHash{$pinName} = 1;
          }
     }
     
PINNAME:     foreach my $pinName (keys(%$pinHashRef)) {
          if($$pinHashRef{$pinName}->{'direction'}[0] =~ /input/) {
               if(!defined($$pinHashRef{$pinName}->{'clk_edge_rate'}[0])) {
                    print_input_timing("",$$pinHashRef{$pinName}->{'clock_pin_true'}[0],1, "", "", $pinName, "", $$pinHashRef{$pinName}->{'pincap'}[0], "", "", "", "", "");
                    delete($tmpHash{$pinName});
               } else {
                    foreach my $arcNum (keys(%$arcHashRef)) {
                         if($$arcHashRef{$arcNum}->{'type'}[0] =~ /setup/) {
                              if($$arcHashRef{$arcNum}->{'input'}[0] =~ /$pinName/) {
                                   if(exists($tmpHash{$pinName})) {
                                        delete($tmpHash{$pinName});
                                        next PINNAME;
                                   } 
                              }
                         }
                    }
               }
          }
     }
     foreach my $foo (keys(%tmpHash)) {
          print_input_timing("",$$pinHashRef{$foo}->{'clock_pin_true'}[0],1, "", "", $foo, "", $$pinHashRef{$foo}->{'pincap'}[0], "", "", "", "", "");
     }
}
